#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
Provide the class to calculate convex hull of a list of 3 dimensional point.
Author: Jinxuan Wu
"""

#from __future__ import division
#import sys


class ConvexHull3D(object):
    def __init__(self, points=[]):
        self.points = points
        self.convexhull = []

    def vect(self, points, start, end):
        """
        Construct an vector
        :param points: input point list
        :param start: start point's index in point list
        :param end: end point's index in point list
        :return: vector from start to end
        """
        return [y - x for x, y in zip(points[start], points[end])]

    def cross(self, u, v):
        """
        Calculate cross product between two vectors
        :return: cross product result
        """
        return [u[1] * v[2] - u[2] * v[1], u[2] * v[0] - u[0] * v[2], u[0] * v[1] - u[1] * v[0]]

    def dot(self, u, v):
        """
        Calculate dot product between two vectors
        :return: dot product result 
        """
        return sum([x * y for x, y in zip(u, v)])

    def normal(self, points, face):
        """
        Return normal vector from face
        :param points: 
        :param face: 
        :return: 
        """
        u = self.vect(points, face[0], face[1])
        v = self.vect(points, face[0], face[-1])
        return self.cross(u, v)

    def seen(self, points, face, p):
        """
        Return if point s is able to see this face or plane
        """
        N = self.normal(points, face)
        P = self.vect(points, face[0], p)
        return (self.dot(N, P) >= 0)

    def bdry(self, faces):
        """
        Filter face that is inside ConvexHull
        :param faces: 
        :return: 
        """
        bdry_fw = set([(F[i - 1], F[i]) for F in faces for i in range(0, len(F))])
        bdry_bk = set([(F[i], F[i - 1]) for F in faces for i in range(0, len(F))])
        return bdry_fw - bdry_bk

    def addPoint(self, points, hull, p):
        """
        Add points to current hull, if the point is not in it. Update by construct faces from point line between seen face
        and un seen faces. 
        :param points: The input point list
        :param hull: Current ConvexHull in which we might need to add point and face to it
        :param p: Point that might need to add
        :return: None.
        """
        seenF = [F for F in hull if self.seen(points, F, p)]
        if len(seenF) == len(hull):  # if can see all faces, unsee ones looking "down"
            N = self.normal(points, seenF[0])
            seenF = [F for F in seenF if self.dot(self.normal(points, F), N) > 0]
        for F in seenF:
            hull.remove(F)
        for E in self.bdry(seenF):
            hull.append([E[0], E[1], p])

    def one_face(self, points, face, p):
        """
        Check if point is on face
        :param points: 
        :param face: 
        :param p: 
        :return: 
        """
        a, b, c = self.normal(points, face)
        d = -(a * points[face[0]][0] + b * points[face[0]][1] + c * points[face[0]][2])
        p_ = a * points[p][0] + b * points[p][1] + c * points[p][2]
        return p_ == d

    def get_convexhull(self):
        """
        Generate ConvexHull from input list.
        :return: 
        """
        return self.hull

    def calc_convexhull(self):
        """
        Generate ConvexHull from input list.
        :return:
        """
        # make sure points are not on same line or same plane. Otherwise we couldn't find build the initial convex hull
        all_in_one_plane = True
        for i in range(3, len(self.points)):
            if not self.one_face(self.points, [0, 1, 2], i):
                all_in_one_plane = False
        if all_in_one_plane:
            self.hull = None
        else:
            hull = [[0, 1, 2], [0, 2, 1]]
            for i in range(3, len(self.points)):
                self.addPoint(self.points, hull, i)
            self.hull = hull


#if __name__ == "__main__":
#    if sys.stdin.isatty():
#        print "Input a integer N, then follows N+K point sets"
#        sys.exit()
#    N = int(sys.stdin.readline())
#    points_str_list = sys.stdin.readlines()
#    all_points = map(lambda x: x.strip().split(), points_str_list)
#    all_points = map(lambda x: [int(y) for y in x], all_points)
#    convexhull_candidate_points = all_points[0:N]
#    print "Construct 3D ConvexHull for a set of " + str(N) + " points"
#    ConvexHull = ConvexHull3D(convexhull_candidate_points)
#    print "Result ConvexHull is: "
#    for point in ConvexHull.get_convexhull():
#        print point
#    for point in all_points[N:]:
#        if ConvexHull.check_if_within_convexhull(point):
#            print "point " + ", ".join([str(x) for x in point]) + " is within the ConvexHull"
#        else:
#            print "point " + ", ".join([str(x) for x in point]) + " is not within the ConvexHull"